/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

window.ScoTools = (function() {
var ScoTools = {};

/** Class of object managing all states of actual SCO */
function ScoStateMgr(
    arrInits        // Optional: Array of functions or objects like { fkt: "function", afterRegistrations: "boolean" } representing functions to call
) {
    this._data = {};
    this._terminationFunctions = [];
    const fkts = (arrInits instanceof Array) ? arrInits : ScoStateMgr.arrInits;
    fkts.filter(obj => (obj instanceof Function) || (obj instanceof Object && !obj.afterRegistrations))
        .forEach(obj => ((obj instanceof Function) ? obj : obj.fkt).call(this));
    fkts.filter(obj => obj instanceof Object && obj.afterRegistrations)
        .forEach(obj => obj.fkt.call(this));
    window.addEventListener('beforeunload', () => {
        this._terminationFunctions.sort((a, b) => a.priority - b.priority);
        this._terminationFunctions.forEach(({ fkt }) => fkt.call(this) );
    });
}

/** What the value to a key is intended to be stored as */
ScoStateMgr.TARGET_NONE = 0;        // Value never needs to be saved (not even if broadcasting)
ScoStateMgr.TARGET_INPUT = 1;       // Value needs only to be saved for broadcasting, but not to restore session
ScoStateMgr.TARGET_DATA = 10;       // Value needs to be saved to restore user's inputed values
ScoStateMgr.TARGET_LOCATION = 11;   // Value needs to be saved to restore basic state of SCO
ScoStateMgr.TARGET_SCORE = 20;      // Value represents raw score to submit
ScoStateMgr.TARGET_MAXSCORE = 21;   // Value represents maximum achievable score
ScoStateMgr.TARGET_STATUS = 22;     // Value represents value of cmi.core.lesson_status

/** Priority of value within its save target group (all integer values are allowed; ignored if save target is lower than ScoStateMgr.TARGET_DATA) */
ScoStateMgr.PRIORITY_INDISPENSABLE = 100;   //    [100] SCO has to be reset to initial state if this value is missing
ScoStateMgr.PRIORITY_NECCASSARY    =  90;   // [81; 99] If not saved very relevant information are missing
ScoStateMgr.PRIORITY_REQUIRED      =  70;   // [61; 80] If not saved a part of the SCO (e.g. rating) is not fully functional
ScoStateMgr.PRIORITY_HOPEDFOR      =  30;   // [ 1; 60] Should be saved, but useage of SCO is possible without this information
ScoStateMgr.PRIORITY_OPTIONAL      =   0;   //      [0] It is nice to save this information even though they are hardly relevant
ScoStateMgr.PRIORITY_NEVER         =  -1;   // ]-∞; -1] Never save this information

/** All information on keys including current value */
ScoStateMgr.prototype._data = {};

/** Functions to call on page termination */
ScoStateMgr.prototype._terminationFunctions = [];

/**
* Adds a new data key
* @return boolean       Added successfully?
*/
ScoStateMgr.prototype.registerKey = function(
    key,                // Internal name of value
    defaultValue,       // Default value if not changed
    validator,          // Optional: Function to transform new value to required type: function(newValue, oldValue) { /* this is ScoStateMgr instance */ return "validated new value"; }
    saveTarget,         // Optional: See ScoStateMgr.TARGET_*
    savePriority,       // Optional: See ScoStateMgr.PRIORITY_*
    latency             // Optional: Timeout in seconds before saving is actually intended (intermediate steps may be skipped)
) {
    if (this._data[key] instanceof Object)
        return false;
    const data = { value: defaultValue, defaultValue, listeners: [] };
    if (validator instanceof Function)
        data.validator = validator;
    if (saveTarget)
        data.saveTarget = saveTarget;
    if (savePriority)
        data.savePriority = savePriority;
    if (latency)
        data.latency = latency;
    this._data[key] = data;
    return true;
}

/**
* Get current value of key `key`
* @return mixed     Current value
*/
ScoStateMgr.prototype.getValue = function(
    key     // Internal name of value
) {
    const data = this._data[key];
    if (data instanceof Object)
        return data.value;
    return undefined;
}

/**
* Set current value of key `key`
* @return undefined
*/
ScoStateMgr.prototype.setValue = function(
    key,    // Internal name of value
    value   // New value
) {
    const data = this._data[key];
    if (!(data instanceof Object))
        return;
    const oldValue = data.value;
    if (data.validator instanceof Function)
        value = data.validator.call(this, value, oldValue);
    if (data.value === value)
        return;
    data.value = value;
    data.listeners.forEach(listener => listener.call(this, value, oldValue));
}

/**
* Add event listener (`function(newValue, oldValue)` beeing executed with ScoStateMgr instance beeing `this`) for change of a specific key
* @return boolean       Was listener addes successfully?
*/
ScoStateMgr.prototype.addListener = function(
    key,        // Internal name of value
    listener,   // Function to call whenever `key`'s value changes
    initialCall // Optional: Set `true` to always enforce a initial call of listener or `false` to suppress the initial call; otherwise `listener` is called if value corresponding to `key` is not the default value
) {
    const data = this._data[key];
    if (!(data instanceof Object))
        return false;
    data.listeners.push(listener);
    if ((initialCall !== false) && ((typeof initialCall == 'boolean') || (data.value !== data.defaultValue)))
        listener.call(this, data.value, data.defaultValue);
    return true;
}

/**
* Get all registered keys
* @return Array     All registered keys (only keys as string values)
*/
ScoStateMgr.prototype.getAllKeys = function() {
    return Object.keys(this._data);
}

/**
* Get save target of key `key`
* @return mixed     Save target (see ScoStateMgr.TARGET_*) or undefined
*/
ScoStateMgr.prototype.getSaveTarget = function(
    key     // Internal name of value
) {
    return (this._data[key] || {}).saveTarget;
}

/**
* Get save priority of key `key`
* @return mixed     Save priority (see ScoStateMgr.PRIORITY_*) or undefined
*/
ScoStateMgr.prototype.getSavePriority = function(
    key     // Internal name of value
) {
    return (this._data[key] || {}).savePriority;
}

/**
* Get default value of key `key`
* @return mixed     Default value set at key registration
*/
ScoStateMgr.prototype.getDefaultValue = function(
    key     // Internal name of value
) {
    return (this._data[key] || {}).defaultValue;
}

/**
* Get latency of key `key`
* @return mixed     Latency in seconds set at key registration
*/
ScoStateMgr.prototype.getLatency = function(
    key     // Internal name of value
) {
    return (this._data[key] || {}).latency || 0;
}

/**
* Register a function to be executed when page is terminated
* @return undefined
*/
ScoStateMgr.prototype.addTerminationFunction = function(
    fkt,        // Function to call on termination
    priority    // Priority of function: Functions with lower priority are executed first
) {
    this._terminationFunctions.push({ fkt, priority });
}

/** Objects like { fkt: "function", afterRegistrations: "boolean" } representing functions to call on object creation */
ScoStateMgr.arrInits = [];

/**
* Add a function to call on constructing a new instance of ScoStateMgr (only if no parameter is given)
* @return undefined
*/
ScoStateMgr.registerInit = function(
    fkt,                // Function to call on new instance of ScoStateMgr
    afterRegistrations  // Optional: Set true if function `fkt` should be called after all registrations
) {
    ScoStateMgr.arrInits.push({ fkt, afterRegistrations });
}



/** Add keys directly linked to SCORM API to ScoStateMgr's keys */
const establishLmslink = (function() {
/**
* Convert Number `v` to string consisting of `countDigits` digits
* @return String    String representation of `v` having required number of digits
*/
function toDigits(
    v,              // Zahl
    countDigits     // Gewünschte Anzahl an Ziffern
) {
    if (v >= Math.pow(10, countDigits))
        v = Math.pow(10, countDigits) - 1;
    if (!Number.isFinite(v))
        v = 0;
    var s = Math.floor(v).toString();
    while (s.length < countDigits)
        s = '0' + s;
    return s;
}

/**
* Convert number of seconds to CMITimespan value
* @return String    Corresponding CMITimespan value
*/
function seconds2timespan(
    seconds     // Number of seconds to be converted
) {
    seconds = Math.round(seconds * 100) / 100;
    if (seconds >= 36000000)
        return '9999:99:99.99';
    return toDigits(seconds / 3600, 4) + ':' + toDigits((seconds / 60) % 60, 2) + ':' + toDigits(seconds % 60, 2) + '.' + toDigits((seconds * 100) % 100, 2);
}

/**
* Convert CMITimespan value to number of seconds
* @return Number        Corresponding number of seconds
*/
function timespan2seconds(
    strTimespan     // CMITimespan value (string like "HHHH:MM:SS[.mm]") to be converted
) {
    if ((typeof strTimespan != 'string') && !(strTimespan instanceof String))
        return 0;
    var a = strTimespan.split(':').map(z => (Number.isFinite(+z) ? (+z) : 0));
    if (a.length > 1)
        a[a.length - 2] *= 60;
    if (a.length > 2)
        a[a.length - 3] *= 3600;
    var v = a.reduce((a, b) => a + b);
    return (v > 36000060) ? Infinity : v;
}

// Register keys
ScoStateMgr && ScoStateMgr.registerInit(function() {
    function validateScore(value) {
        if ((value === '') || !(('' + value).length))
            return '';
        const v = Math.max(0, Math.min(+value, 100));
        return Number.isFinite(v) ? v : '';
    }

    this.registerKey('connection.lms', false, (value, old) => (value || old) ? true : false);
    this.registerKey('lms.studentName', '', (value, old) => (old || value));
    this.registerKey('lms.lessonLocation', '', (value, old) => {
        value = '' + value;
        if (value.split('').reduce((prev, curr) => prev || (curr.charCodeAt(0) > 127), false))
            return old;
        if (value.length > 255)
            return old;
        return value;
    });
    this.registerKey('lms.lessonStatus', 'not attempted', value => (['passed', 'completed', 'failed', 'incomplete', 'browsed'].indexOf(value) == -1) ? 'not attempted' : value);
    this.registerKey('lms.entry', '', value => (['ab-initio', 'resume'].indexOf(value) == -1) ? '' : value);
    this.registerKey('lms.score', '', validateScore);
    this.registerKey('lms.scoreMin', '', validateScore);
    this.registerKey('lms.scoreMax', '', validateScore);
    this.registerKey('lms.totalTime', 0, value => Number.isFinite(+value) ? Math.max(0, +value) : 0);
    this.registerKey('lms.sessionTime', 0, value => Number.isFinite(+value) ? Math.max(0, +value) : 0);
    this.registerKey('lms.lessonMode', 'normal', value => (['browse', 'review'].indexOf(value) == -1) ? 'normal' : value);
    this.registerKey('lms.suspendData', '', (value, old) => {
        value = '' + value;
        if (value.split('').reduce((prev, curr) => prev || (curr.charCodeAt(0) > 127), false))
            return old;
        if (value.length > 4096)
            return old;
        return value;
    });
    this.registerKey('lms.audio', 0, value => Number.isFinite(+value) ? Math.max(-1, Math.min(100, Math.round(+value))) : 0);
    this.registerKey('lms.speed', 0, value => Number.isFinite(+value) ? Math.max(-100, Math.min(100, Math.round(+value))) : 0);
    this.registerKey('lms.text', 0, value => Number.isFinite(+value) ? Math.max(-1, Math.min(1, Math.round(+value))) : 0);
    this.registerKey('lms.exit', '', (value, old) => (['time-out', 'suspend', 'logout', ''].indexOf(value) == -1) ? old : value);
});
ScoStateMgr && ScoStateMgr.registerInit(function() {
    // Find API
    const api = (function() {
        for (let w = window; w; w = (w.parent == w) ? null : w.parent) {
            if (w.API && (typeof w.API.LMSInitialize == "function"))
                return w.API;
        }
        return null;
    })();
    if (!api || api.LMSInitialize("") != "true")
        return;
    // Read values from LMS
    [
        { from: 'cmi.core.student_name', to: 'lms.studentName' },
        { from: 'cmi.core.lesson_location', to: 'lms.lessonLocation' },
        { from: 'cmi.core.lesson_status', to: 'lms.lessonStatus' },
        { from: 'cmi.core.entry', to: 'lms.entry' },
        { from: 'cmi.core.score.min', to: 'lms.scoreMin' },
        { from: 'cmi.core.score.max', to: 'lms.scoreMax' },
        { from: 'cmi.core.score.raw', to: 'lms.score' },
        { from: 'cmi.core.lesson_mode', to: 'lms.lessonMode' },
        { from: 'cmi.suspend_data', to: 'lms.suspendData' },
        { from: 'cmi.student_preference.audio', to: 'lms.audio' },
        { from: 'cmi.student_preference.speed', to: 'lms.speed' },
        { from: 'cmi.student_preference.text', to: 'lms.text' }
    ].forEach(({ from, to }) => this.setValue(to, api.LMSGetValue(from)));
    this.setValue('cmi.core.exit', '');
    this.setValue('lms.totalTime', timespan2seconds(api.LMSGetValue('cmi.core.total_time')));
    this.setValue('connection.lms', true);
    // Listen on change of values
    let changedKeys = [];
    let timeoutNormal = undefined;
    let timeoutDelayed = undefined;
    const writeChanges = () => {
        timeoutNormal = undefined;
        timeoutDelayed = undefined;
        if (!changedKeys.length)
            return;
        while (changedKeys.length) {
            const key = changedKeys.shift();
            switch (key) {
                case 'lms.lessonLocation':
                    api.LMSSetValue('cmi.core.lesson_location', this.getValue(key));
                    break;
                case 'lms.lessonStatus':
                    api.LMSSetValue('cmi.core.lesson_status', this.getValue(key));
                    break;
                case 'lms.suspendData':
                    api.LMSSetValue('cmi.suspend_data', this.getValue(key));
                    break;
                case 'lms.score':
                    const scoreRaw = this.getValue(key);
                    const scoreMin = this.getValue('lms.scoreMin');
                    if (scoreMin === '' && scoreRaw !== '') {
                        this.setValue('lms.scoreMin', 0);
                        api.LMSSetValue('cmi.core.score.min', '0');
                    } else
                        api.LMSSetValue('cmi.core.score.min', '' + scoreMin);

                    const scoreMax = this.getValue('lms.scoreMax');
                    if (scoreMax === '' && scoreRaw !== '') {
                        this.setValue('lms.scoreMax', 100);
                        api.LMSSetValue('cmi.core.score.max', '100');
                    } else
                        api.LMSSetValue('cmi.core.score.max', '' + scoreMax);

                    api.LMSSetValue('cmi.core.score.raw', '' + scoreRaw);
                    break;
                case 'lms.audio':
                    api.LMSSetValue('cmi.student_preference.audio', this.getValue(key));
                    break;
                case 'lms.speed':
                    api.LMSSetValue('cmi.student_preference.speed', this.getValue(key));
                    break;
                case 'lms.text':
                    api.LMSSetValue('cmi.student_preference.text', this.getValue(key));
                    break;
                case 'lms.exit':
                    api.LMSSetValue('cmi.core.exit', this.getValue(key));
                    break;
                case 'lms.sessionTime':
                    api.LMSSetValue('cmi.core.session_time', seconds2timespan(this.getValue(key)));
                    break;
            }
        }
        api.LMSCommit('');
    };
    ['lms.lessonLocation', 'lms.lessonStatus', 'lms.score', 'lms.suspendData', 'lms.audio', 'lms.speed', 'lms.text', 'lms.exit']
        .forEach(key => this.addListener(key, () => {
            if (changedKeys.indexOf(key) == -1)
                changedKeys.push(key);
            if (timeoutDelayed !== undefined) {
                clearTimeout(timeoutDelayed);
                timeoutDelayed = undefined;
            }
            if (timeoutNormal === undefined)
                timeoutNormal = setTimeout(writeChanges, 0);
        }, false));
    this.addListener('lms.sessionTime', () => {
        if (changedKeys.indexOf('lms.sessionTime') == -1)
            changedKeys.push('lms.sessionTime');
        if (timeoutNormal === undefined && timeoutDelayed === undefined)
            timeoutDelayed = setTimeout(writeChanges, 30000);
    });
    this.addTerminationFunction(() => {
        writeChanges();
        api.LMSFinish('');
    }, Infinity);
}, true);
});
establishLmslink();



/** Pass (and pack/unpack) values gotten from LMS (via corresponding ScoStateMgr's keys) to content's keys and vice versa  */
(function(){

/**
* Convert normal UTF-16 JavaScript string to string only containing ASCII characters in a space efficient way
* @return String        Converted version of `str`
*/
function utf16ToAscii(
    str         // UTF16 string to convert
) {
    let encoded = '';
    for (let n = 0; n < str.length; ++n) {
        let code = str.charCodeAt(n);
        if (code == 10)
            code = 127;
        else if (code == 127)
            code = 10;
        switch (true) {
            case (code < 32):
                encoded += '\u001b' + String.fromCharCode(127 - code);
                break;
            case (code < 128):
                encoded += String.fromCharCode(code);
                break;
            case (code < 28 * 128 - 33 + 128):
                code -= 128;
                encoded += String.fromCharCode(Math.floor(code / 128)) + String.fromCharCode(code & 127);
                break;
            default:
                encoded += String.fromCharCode(28 + ((code & 32768) ? 2 : 0) + ((code & 16384) ? 1 : 0))
                    + String.fromCharCode(Math.floor(code / 128) & 127)
                    + String.fromCharCode(code & 127);
        }
    }
    return encoded;
}

/**
* Convert encoding of normal UTF-16 JavaScript string as string only containing ASCII characters back to encoded string
* @return String        Decoded version of `str`
*/
function asciiToUtf16(
    str         // String only containing ASCII characters
) {
    let decoded = '';
    for (let n = 0; n < str.length; ++n) {
        let code = str.charCodeAt(n) & 127;
        switch (true) {
            case (code == 127):
                code = 10;
                break;
            case (code > 31):
                break;
            case (code > 27):
                code = (code - 28) * 16384 + str.charCodeAt(++n) * 128 + str.charCodeAt(++n);
                break;
            case (code == 27):
                const next = str.charCodeAt(n + 1);
                if (next > 95) {
                    ++n;
                    code = 127 - next;
                    if (code == 10)
                        code = 127;
                    break;
                }
                // nobreak
            default:
                code = code * 128 + str.charCodeAt(++n) + 128;
                break;
        }
        decoded += String.fromCharCode(code);
    }
    return decoded;
}

// Register keys
ScoStateMgr && ScoStateMgr.registerInit(function() {
    this.registerKey('pool4lms.active', false, value => value ? true : false);
    this.registerKey('pool4lms.unsatisfied', -1, value => (value == -1) ? -1 : Math.max(0, Math.min(100, Math.round(value))));
});

// Prepaire pooling
ScoStateMgr && ScoStateMgr.registerInit(function() {
    // What keys should be saved where?
    const relevantKeys = {};
    this.getAllKeys().forEach(key => {
        const target = this.getSaveTarget(key);
        if (!(target >= 10))
            return;
        if (!relevantKeys[target])
            relevantKeys[target] = [];
        relevantKeys[target].push({ key, priority: this.getSavePriority(key) });
    });
    Object.keys(relevantKeys).forEach(target => {
        relevantKeys[target].sort((a, b) => (a.priority == b.priority) ? a.key.localeCompare(b.key) : (b.priority - a.priority));
        if (target >= 20)
            relevantKeys[target] = relevantKeys[target][0].key;
    });

    // Functions to write values to lms.*
    const unsatisfied = new Array(10);
    const calcUnsatisfied = () => this.setValue('pool4lms.unsatisfied', unsatisfied.reduce((a, b) => Math.max(a, b)));
    const packValue = (target, noSave) => {
        let pack = '';
        let minPriority = 0;
        let freeSpace, saveKey;
        switch (target) {
            case ScoStateMgr.TARGET_DATA:
                freeSpace = 4096;
                saveKey = 'lms.suspendData';
                break;
            case ScoStateMgr.TARGET_LOCATION:
                freeSpace = 255;
                saveKey = 'lms.lessonLocation';
                break;
            default:
                return;
        }
        for (let n = 0; n < relevantKeys[target].length; ++n) {
            const value = this.getValue(relevantKeys[target][n].key);
            const s = (value == this.getDefaultValue(relevantKeys[target][n].key)) ? '' : ((value === undefined) ? 'undefined' : utf16ToAscii(JSON.stringify(value)));
            const sSpace = s.length;
            if (sSpace < freeSpace) {
                pack += (pack.length ? ',' : '[') + s;
                freeSpace -= sSpace + 1;
            } else {
                minPriority = relevantKeys[target][n].priority;
                break;
            }
        }
        if ((pack.length || !minPriority) && !noSave)
            this.setValue(saveKey, pack);
        unsatisfied[target - 10] = minPriority;
        calcUnsatisfied();
    };
    const passValue = (target) => {
        const key = relevantKeys[target];
        if (key === undefined)
            return;
        const value = this.getValue(key);
        switch(target) {
            case ScoStateMgr.TARGET_STATUS:
                this.setValue('lms.lessonStatus', value);
                break;
            case ScoStateMgr.TARGET_SCORE:
                this.setValue('lms.score', value);
                break;
            case ScoStateMgr.TARGET_MAXSCORE:
                this.setValue('lms.scoreMax', value);
                break;
        }
    };

    // Save multiple values
    const changedTargets = [];
    let timeoutDelayed = undefined;
    let timestampNextSave = Infinity;
    const saveMultiple = () => {
        while (changedTargets.length) {
            const target = changedTargets.shift();
            (target < 20) ? packValue(target) : passValue(target);
        }
        timeoutDelayed = undefined;
        timestampNextSave = Infinity;

    };

    // Add another target to list of changed targets and set timer
    const prepareSave = (key, target) => {
        if (!this.getValue('pool4lms.active'))
            return;
        if (changedTargets.indexOf(target) == -1)
            changedTargets.push(target);
        const latency = this.getLatency(key) * 1000;
        const wantedSaveTime = (new Date()).getTime() + latency;
        if (timestampNextSave <= wantedSaveTime)
            return;
        if (timeoutDelayed !== undefined)
            clearTimeout(timeoutDelayed);
        timeoutDelayed = setTimeout(saveMultiple, latency);
        timestampNextSave = wantedSaveTime;
        if ((target < 20) && latency)
            packValue(target, true);
    };

    // React on change of a value having a save target specified
    Object.keys(relevantKeys).forEach(target => {
        const keys = relevantKeys[+target];
        if (target >= 20) {
            this.addListener(keys, value => prepareSave(keys, +target));
            return;
        }
        keys.forEach(({ key }) => this.addListener(key, value => prepareSave(key, +target)));
    });

    // Get corresponding lms.* key
    const getLmsKey = target => {
        switch (+target) {
            case ScoStateMgr.TARGET_DATA:
                return 'lms.suspendData';
            case ScoStateMgr.TARGET_LOCATION:
                return 'lms.lessonLocation';
            case ScoStateMgr.TARGET_STATUS:
                return 'lms.lessonStatus';
            case ScoStateMgr.TARGET_SCORE:
                return 'lms.score';
            case ScoStateMgr.TARGET_MAXSCORE:
                return 'lms.scoreMax';
            default:
                return undefined;
        }
    };

    // Split string packing values to single value representing strings
    const splitPacked = s => {
        s = '' + s;
        if (s.charAt(0) != '[')
            return [];
        const result = [];
        const toClose = [];
        let posStart = 1;
        let posEnd = 1;
        for (posEnd; posEnd < s.length; ++posEnd) {
            const c = s.charAt(posEnd);
            if (c == '\\') {
                ++posEnd;
                continue;
            }
            if (toClose.length && c == toClose[toClose.length -1]) {
                toClose.pop();
                continue;
            }
            switch (c) {
                case "'":
                case '"':
                    toClose.push(c);
                    continue;
                case '[':
                    toClose.push(']');
                    continue;
                case '(':
                    toClose.push(')');
                    continue;
                case '{':
                    toClose.push('}');
                    continue;
            }
            if (!toClose.length && c == ',') {
                result.push(s.substring(posStart, posEnd));
                posStart = posEnd + 1;
            }
        }
        result.push(s.substring(posStart, posEnd));
        return result.map(s => asciiToUtf16(s));
    };

    // Read values from corresponding lms.* value
    const recoverValues = (target) => {
        let lmskey = getLmsKey(target);
        if (!lmskey || !relevantKeys[target])
            return;
        if (target >= 20) {
            this.setValue(relevantKeys[target], this.getValue(lmskey));
            return;
        }
        splitPacked(this.getValue(lmskey)).forEach((value, index) => {
            if (index >= relevantKeys[target].length)
                return;
            const key = relevantKeys[target][index].key;
            if (value === '') {
                this.setValue(key, this.getDefaultValue(key));
                return;
            }
            if (value === 'undefined') {
                this.setValue(key, undefined);
                return;
            }
            try {
                this.setValue(key, JSON.parse(value));
            } catch(e) {}
        });
    };

    // Read values previously saved by lms
    Object.keys(relevantKeys).forEach(target => {
        recoverValues(target);
        this.addListener(getLmsKey(target), () => recoverValues(target));
    });

    // React on activation or deactivation of pooling
    this.addListener('pool4lms.active', active => {
        if (!active) {
            this.setValue('pool4lms.unsatisfied', -1);
            return;
        }
        Object.keys(relevantKeys).forEach(target => (target < 20) ? packValue(target) : passValue(target));
    });

    // Set `lms.exit` and eventually `lms.lessonStatus` when `connection.lms` changes to `true`
    this.addListener('connection.lms', active => {
        if (!active)
            return;
        this.setValue('lms.exit', 'suspend');
        if (this.getValue('lms.lessonStatus') == 'not attempted')
            this.setValue('lms.lessonStatus', 'incomplete');
    });

    /** Recalculate if pooling is neccessary */
    const recalculateActive = () => {
        this.setValue(
            'pool4lms.active',
            this.getValue('connection.lms')
                && this.getValue('lms.lessonMode') == 'normal'
                && (['not attempted', 'incomplete', 'browsed'].indexOf(this.getValue('lms.lessonStatus')) > -1)
        );
    };
    this.addListener('connection.lms', recalculateActive);
    this.addListener('lms.lessonStatus', recalculateActive);
    recalculateActive();

}, true);
})();



/** Measure time student attends SCO and write meassured time to lms.sessionTime */
(function() {
// Register keys
ScoStateMgr && ScoStateMgr.registerInit(function() {
    function validateTill(value, oldValue) {
        if (typeof value == 'string')
            value = +value;
        if (Number.isFinite(value)) {
            const end = (new Date()).getTime() + value * 1000;
            if (end - 250 <= ((oldValue instanceof Date) ? oldValue.getTime() : oldValue))
                return oldValue;
            value = new Date(end);
        }
        if (value == Infinity)
            return Infinity;
        if (value instanceof Date && value > new Date())
            return value;
        return null;
    }

    this.registerKey('stopwatch.isActiveAttendance', true, value => value ? true : false);
    this.registerKey('stopwatch.stdMaxCountActive', Infinity, value => Number.isFinite(+value) ? Math.max(0, +value) : Infinity);
    this.registerKey('stopwatch.stdMaxCountInactive', 5, value => Number.isFinite(+value) ? Math.max(0, +value) : Infinity);
    this.registerKey('stopwatch.maxCountActiveTill', null, validateTill);
    this.registerKey('stopwatch.maxCountInactiveTill', null, validateTill);
});

// Enlive meassurement
ScoStateMgr && ScoStateMgr.registerInit(function() {
    let lastTimestamp = new Date();
    const updateSessionTime = () => {
        const lastUpdated = lastTimestamp;
        const now = new Date();
        lastTimestamp = now;
        const maxCountTill = this.getValue(this.getValue('stopwatch.isActiveAttendance') ? 'stopwatch.maxCountActiveTill' : 'stopwatch.maxCountInactiveTill');
        if (maxCountTill === null)
            return;
        const extraCount = Math.max(0, ((maxCountTill > now) ? now : maxCountTill).getTime() - lastUpdated.getTime());
        if (extraCount)
            this.setValue('lms.sessionTime', this.getValue('lms.sessionTime') + extraCount / 1000);
    };

    const endPhase = (forActivePhase, maxCountTill) => {
        const lastUpdated = lastTimestamp;
        const now = new Date();
        lastTimestamp = now;
        if (forActivePhase !== undefined && this.getValue('stopwatch.isActiveAttendance') != forActivePhase)
            return;
        if (maxCountTill === null)
            return;
        const extraCount = Math.max(0, ((maxCountTill > now) ? now : maxCountTill).getTime() - lastUpdated.getTime());
        if (extraCount)
            this.setValue('lms.sessionTime', this.getValue('lms.sessionTime') + extraCount / 1000);
    };

    const updateMaxCountActiveTill = () => this.setValue('stopwatch.maxCountActiveTill', this.getValue('stopwatch.stdMaxCountActive'));
    const updateMaxCountInactiveTill = () => this.setValue('stopwatch.maxCountInactiveTill', this.getValue('stopwatch.stdMaxCountInactive'));

    const updateIsActiveAttendance = () => this.setValue('stopwatch.isActiveAttendance', document.visibilityState === "visible" && document.hasFocus());
    const loopUpdate = () => { updateIsActiveAttendance(); window.requestAnimationFrame(loopUpdate); };
    window.addEventListener("focus", updateIsActiveAttendance);
    window.addEventListener("blur", updateIsActiveAttendance);
    document.addEventListener("visibilitychange", updateIsActiveAttendance);

    this.addListener('stopwatch.isActiveAttendance', (value, oldValue) => {
        endPhase(undefined, this.getValue(oldValue ? 'stopwatch.maxCountActiveTill' : 'stopwatch.maxCountInactiveTill'));
        value ? updateMaxCountActiveTill() : updateMaxCountInactiveTill();
    });
    this.addListener('stopwatch.stdMaxCountActive', () => this.setValue('stopwatch.maxCountActiveTill', 'stopwatch.stdMaxCountActive'));
    this.addListener('stopwatch.stdMaxCountInactive', () => this.setValue('stopwatch.maxCountInactiveTill', 'stopwatch.stdMaxCountInactive'));
    this.addListener('stopwatch.maxCountActiveTill', (newValue, oldValue) => endPhase(true, oldValue));
    this.addListener('stopwatch.maxCountInactiveTill', (newValue, oldValue) => endPhase(false, oldValue));

    updateMaxCountActiveTill();
    loopUpdate();
    window.setInterval(updateSessionTime, 15000);
    this.addTerminationFunction(updateSessionTime, 0);
});
})();



/** Awards input events with seconds messureable by stopwatch */
(function() {
// Register keys
ScoStateMgr && ScoStateMgr.registerInit(function() {
    function validateSeconds(value) {
        return Number.isFinite(+value) ? Math.max(0, +value) : 0;
    }
    this.registerKey('stopwatch.countScrollActive', 10, validateSeconds);
    this.registerKey('stopwatch.countScrollInactive', 10, validateSeconds);
    this.registerKey('stopwatch.countPointerActive', 20, validateSeconds);
    this.registerKey('stopwatch.countPointerInactive', 10, validateSeconds);
    this.registerKey('stopwatch.countInput', 30, validateSeconds);
});

// Apply times for stopwatch messurements whenever corresponding events occure (only if that increases meassureable time)
ScoStateMgr && ScoStateMgr.registerInit(function() {
    const addSeconds = (forActivePhase, keySeconds) => {
        let seconds = this.getValue(keySeconds);
        if (!seconds)
            return;
        const key = forActivePhase ? 'stopwatch.maxCountActiveTill' : 'stopwatch.maxCountInactiveTill';
        this.setValue(key, seconds);
    }
    document.addEventListener('scroll', () => {
        this.getValue('stopwatch.isActiveAttendance') ? addSeconds(true, 'stopwatch.countScrollActive') : addSeconds(false, 'stopwatch.countScrollInactive');
    });
    ['input', 'keypress'].forEach(eventname => document.addEventListener(eventname, () => {
        addSeconds(this.getValue('stopwatch.isActiveAttendance'), 'stopwatch.countInput');
    }));
    ['mousemove', 'touchmove', 'pointermove'].forEach(eventname => document.addEventListener(eventname, () => {
        this.getValue('stopwatch.isActiveAttendance') ? addSeconds(true, 'stopwatch.countPointerActive') : addSeconds(false, 'stopwatch.countPointerInactive');
    }));
}, true);
})();



/** Send or receive values over postMessage interface */
(function() {
// Register keys
ScoStateMgr && ScoStateMgr.registerInit(function() {
    function validateBoolean(value) {
        return value ? true : false;
    }
    this.registerKey('connection.broadcast', false, validateBoolean);
    this.registerKey('broadcast.isSender', false, validateBoolean);
    this.registerKey('broadcast.isOnlyDisplay', false, validateBoolean);
    this.registerKey('broadcast.isPresentator', false, validateBoolean);
});

// Start communication if available
ScoStateMgr && ScoStateMgr.registerInit(function() {
    const PROTOCOL_ID = 'de.roith-unterrichtet.remote';
    const relevantKeys = [];
    const changedKeys = [];
    let timeoutId = undefined;

    // Send values of keys given as single parameter (defaults to `changedKeys` if parameter is falsy; in this case or if parameter matches `changedKeys`, Array is set to length 0)
    const sendKeys = keys => {
        timeoutId = undefined;
        keys = keys || changedKeys;
        if (!keys.length || !this.getValue('broadcast.isSender'))
            return;
        let message = PROTOCOL_ID;
        try {
            keys.forEach(key => message += '\n' + key + '=' + JSON.stringify(this.getValue(key)));
            window.parent.postMessage(message, '*');
        } catch(e) {}
        if (keys == changedKeys)
            changedKeys.length = 0;
    };

    // Find (new) relevant keys and listen to their changes
    const updateRelevantKeys = () => {
        this.getAllKeys().forEach(key => {
            if (!(this.getSaveTarget(key) >= ScoStateMgr.TARGET_INPUT))
                return;
            if (relevantKeys.indexOf(key) != -1)
                return;
            relevantKeys.push(key);
            this.addListener(key, () => {
                if (!this.getValue('broadcast.isSender'))
                    return;
                if (changedKeys.indexOf(key) == -1)
                    changedKeys.push(key);
                if (timeoutId === undefined)
                    timeoutId = setTimeout(sendKeys, 0);
            });
        });
    };
    this.addListener('connection.broadcast', value => value && updateRelevantKeys());
    this.addListener('broadcast.isSender', value => value && sendKeys(relevantKeys));

    // Listen to incoming messages
    window.addEventListener('message', ({ data }) => {
        if (!(typeof data == 'string') && !(data instanceof String))
            return;
        data = data.split('\n');
        if (data[0] != PROTOCOL_ID)
            return;
        this.setValue('connection.broadcast', true);
        data.shift();
        data.forEach(item => {
            const indexSeperator = item.indexOf('=');
            if (indexSeperator < 0)
                return;
            try {
                const key = item.substring(0, indexSeperator);
                if (key == 'role') {
                    switch (item.substring(indexSeperator + 1)) {
                        case 'receiver':
                            this.setValue('broadcast.isOnlyDisplay', true);
                            this.setValue('broadcast.isSender', false);
                            this.setValue('broadcast.isPresentator', false);
                            return;
                        case 'sender':
                            this.setValue('broadcast.isOnlyDisplay', false);
                            this.setValue('broadcast.isSender', true);
                            this.setValue('broadcast.isPresentator', false);
                            return;
                        case 'presentator':
                            this.setValue('broadcast.isOnlyDisplay', false);
                            this.setValue('broadcast.isSender', true);
                            this.setValue('broadcast.isPresentator', true);
                            return;
                        case 'normal':
                        default:
                            this.setValue('broadcast.isOnlyDisplay', false);
                            this.setValue('broadcast.isSender', false);
                            this.setValue('broadcast.isPresentator', false);
                            return;
                    }
                }
                if (relevantKeys.indexOf(key) < 0)
                    return;
                const value = JSON.parse(item.substring(indexSeperator + 1));
                this.setValue(key, value);
            } catch(e) {}
        });
    });

    // Send initial message to parent window
    try {
        if (window.parent != window)
            window.parent.postMessage(PROTOCOL_ID, '*');
    } catch(e) {}
}, true);
})();

/**
* Element mit angegebenem Namespace in DOM-Baum einfügen
* @return Element   Erzeugter Knoten
*/
function createNS(
    strNamespace,   // Namespace des ggf zu erzeugenden Elements
    strTagname,     // Optional: Tagname des zu erzeugenden Elementes
    nodeParent,     // Optional: Künftiger Elternknoten
    objAtts,        // Optional: Attribute des zu erzeugenden Elementes
    strTextContent, // Optional: Textinhalt des Elementes
    nodeBefore      // Optional: Knoten, als deren previousSibling der neue Knoten eingefügt werden soll
) {
    var node = nodeBefore ? nodeBefore.parentNode : nodeParent;
    if (nodeBefore && nodeBefore.parentNode)
        nodeParent = nodeBefore.parentNode;
    var doc = nodeParent ? nodeParent.ownerDocument : document;
    if (strTagname)
        node = doc.createElementNS(strNamespace, strTagname);
    if (objAtts)
        for (var key in objAtts)
            node.setAttribute(key, objAtts[key]);
    if ((strTextContent !== undefined) && ('' + strTextContent).length) {
        var nodeText = doc.createTextNode(strTextContent);
        if (node != nodeParent)
            node.appendChild(nodeText);
        else
            node = nodeText;
    }
    if (nodeParent && (node != nodeParent))
        nodeParent.insertBefore(node, nodeBefore);
    return node;
}

/**
* HTML-Element in DOM-Baum einfügen
* @return Element   Erzeugter Knoten
*/
function createHtml(
    strTagname,     // Optional: Tagname des zu erzeugenden Elementes
    nodeParent,     // Optional: Künftiger Elternknoten
    objAtts,        // Optional: Attribute des zu erzeugenden Elementes
    strTextContent, // Optional: Textinhalt des Elementes
    nodeBefore      // Optional: Knoten, als deren previousSibling der neue Knoten eingefügt werden soll
) {
    return createNS('htt' + 'p://w' + 'ww.w3.org/1999/xhtml', strTagname, nodeParent, objAtts, strTextContent, nodeBefore);
}

/**
* SVG-Element in DOM-Baum einfügen
* @return Element   Erzeugter Knoten
*/
function createSvg(
    strTagname,     // Optional: Tagname des zu erzeugenden Elementes
    nodeParent,     // Optional: Künftiger Elternknoten
    objAtts,        // Optional: Attribute des zu erzeugenden Elementes
    strTextContent, // Optional: Textinhalt des Elementes
    nodeBefore      // Optional: Knoten, als deren previousSibling der neue Knoten eingefügt werden soll
) {
    return createNS('htt' + 'p://w' + 'ww.w3.org/2000/svg', strTagname, nodeParent, objAtts, strTextContent, nodeBefore);
}

/**
* Klasse, die eine Seite modal anzeigt
*/
function ModalPage(
    href,       // URL der anzuzeigenden Seite
    closeLabel  // Optional: Text auf Schließen-Schaltfläche
) {
    Object.defineProperty(this, 'node', {
        enumerable: true,
        writable: false,
        value: createHtml('div', null, { 'class': 'modal' })
    });
    Object.defineProperty(this, 'iframe', {
        enumerable: true,
        writable: false,
        value: createHtml('iframe', createHtml('div', createHtml('div', this.node)), { 'src': href })
    });
    var iframe = this.iframe;
    iframe.style.height = '33vh';
    iframe.addEventListener('load', function() {
        try {
            iframe.style.height = Math.min((iframe.contentDocument ? iframe.contentDocument : iframe.contentWindow.document).body.parentNode.scrollHeight + 10, window.innerHeight - 32) + 'px';
        } catch(ex) {}
    });
    Object.defineProperty(this, 'closeBtn', {
        enumerable: true,
        writable: false,
        value: createHtml('a', createHtml('p', this.iframe.parentNode), { 'href': '#' }, closeLabel ? closeLabel : 'Infokasten schließen')
    });
    var self = this;
    this.closeBtn.addEventListener('click', function(ev) {
        ev.preventDefault();
        ev.stopPropagation();
        self.close();
    });
    this.node.addEventListener('click', function() { self.close() ; });
}

/**
* Modales Fenster anzeigen
* @return undefined
*/
ModalPage.prototype.open = function() {
    document.body.appendChild(this.node);
}

/**
* Modales Fenster ggf. schließen
* @return undefined
*/
ModalPage.prototype.close = function() {
    if (this.node.parentNode)
        this.node.parentNode.removeChild(this.node);
}

/**
* Klasse, die einen Link zu einer modal anzuzeigenden Seite mit weiteren Infos verwaltet
*/
function InfoPageLink(
    node        // Optional: HTMLAnchorElement, welches eine modale Seite öffnen soll
) {
    if (!node) {
        (node = document.getElementsByTagName('footer')[0]) && (node = node.lastChild);
    }
    if (node)
        node.addEventListener('click', function(ev) {
            ev.preventDefault();
            ev.stopPropagation();
            (new ModalPage(this.href)).open();
        });
}

/**
* Eine anklickbare Grafik, die Aktionen auslöst
*/
function ImageLink(
    hint,       // Hinweistext
    src,        // URL der Grafik
    alt,        // Optional: Alternativtext (Standardwert: hint)
    onclick     // Optional: Funktion, die beim Anklicken ausgeführt werden soll
) {
    this.node = ImageLink_createA(this);
    createHtml('img', this.node, { 'src': src, 'alt': alt ? alt : hint });
    this.hint = hint;
    if (onclick)
        this.onclick = onclick;
}

/**
* Neues A-Element erstellen
* @return HTMLAnchorElement
*/
function ImageLink_createA(
    self        // Zuständige ImageLink-Instanz
) {
    var node = createHtml('a');
    node.setAttribute('href', '#');
    node.setAttribute('role', 'button');
    node.className = 'ImageLink';
    node.addEventListener('click', function(ev) {
        ev.preventDefault();
        ev.stopPropagation();
        self.click();
    });
    return node;
}

// Aktuelles HTML-Element
ImageLink.prototype.node = null;

// Deaktiviert?
Object.defineProperty(ImageLink.prototype, 'disabled', {
    enumerable: true,
    get() {
        return (this.node.nodeName.toUpperCase() != 'A');
    },
    set(v) {
        var node = null;
        if (this.disabled) {
            if (v)
                return;
            node = ImageLink_createA(this);
        } else {
            if (!v)
                return;
            node = createHtml('span', null, { 'class': 'ImageLink', 'role': 'button', 'aria-disabled': 'aria-disabled' });
        }
        if (this.node.hasAttribute('title'))
            node.setAttribute('title', this.node.getAttribute('title'));
        node.appendChild(this.node.firstChild);
        if (this.node.parentNode)
            this.node.parentNode.replaceChild(node, this.node);
        this.node = node;
    }
});

// URL der Grafik
Object.defineProperty(ImageLink.prototype, 'src', {
    enumerable: true,
    get() {
        return this.node.firstChild.src;
    },
    set(s) {
        this.node.firstChild.src = s;
    }
});

// Alternativtext der Grafik
Object.defineProperty(ImageLink.prototype, 'alt', {
    enumerable: true,
    get() {
        return this.node.firstChild.alt;
    },
    set(s) {
        this.node.firstChild.alt = s;
    }
});

// Hinweistext
Object.defineProperty(ImageLink.prototype, 'hint', {
    enumerable: true,
    get() {
        return (this.node.getAttribute('title') || '');
    },
    set(s) {
        if (s)
            this.node.setAttribute('title', s);
        else
            this.node.removeAttribute('title');
    }
});

// Was beim Anklicken passieren soll
ImageLink.prototype.onclick = null;

/**
* Auf Anklicken reagieren
* @return undefined
*/
ImageLink.prototype.click = function() {
    if (typeof this.onclick == 'function')
        this.onclick.call(this);
}

/**
* Die ImageLink-Instanz old durch this ersetzen
* @return undefined
*/
ImageLink.prototype.replace = function(
    old     // ImageLink-Instanz, die ersetzt werden soll
) {
    if ((old instanceof ImageLink) && (old != this) && (old.node.parentNode))
        old.node.parentNode.replaceChild(this.node, old.node);
}

/**
* Diesen Link als Kind von parentNode im DOM-Baum einfügen
* @return undefined
*/
ImageLink.prototype.appendAsChild = function(
    parentNode      // Neuer Elternknoten
) {
    parentNode.appendChild(this.node);
}

/**
* Eine Meldung, die dem Benutzer angezeigt wird
*/
function Message(
    text,       // Meldungstext
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    Object.defineProperty(this, 'node', {
        enumerable: true,
        value: createHtml('p')
    });
    createHtml('span', createHtml('span', this.node));
    this.type = type;
    if (text)
        this.text = text;
}

// Mögliche Meldungsarten
Object.defineProperty(Message, 'TYPE_SUCCESS', { enumerable: true, value: 'msg_success' });
Object.defineProperty(Message, 'TYPE_FAILED', { enumerable: true, value: 'msg_failed' });
Object.defineProperty(Message, 'TYPE_WARN', { enumerable: true, value: 'msg_warn' });
Object.defineProperty(Message, 'TYPE_INFO', { enumerable: true, value: 'msg_info' });
Object.defineProperty(Message, 'TYPES', { enumerable: true, value: [ Message.TYPE_FAILED, Message.TYPE_SUCCESS, Message.TYPE_INFO, Message.TYPE_WARN ] });

// Meldungsart
Object.defineProperty(Message.prototype, 'type', {
    enumerable: true,
    get() {
        return this.node.className;
    },
    set(s) {
        if (Message.TYPES.indexOf(s) == -1)
            s = s ? Message.TYPE_SUCCESS : Message.TYPE_FAILED;
        this.node.className = s;
    }
});

// Meldungstext
Object.defineProperty(Message.prototype, 'text', {
    enumerable: true,
    get() {
        return this.node.firstChild.firstChild.textContent;
    },
    set(v) {
        this.node.firstChild.firstChild.textContent = v;
    }
});

// Genutzte MessagePlace-Instanz
Message.prototype._place = null;
Object.defineProperty(Message.prototype, 'place', {
    enumerable: true,
    get() {
        return this._place;
    },
    set(place) {
        this.addTo(place);
    }
});

/**
* Diese Meldung auf MessagePlace-Instanz place anzeigen
* @return undefined
*/
Message.prototype.addTo = function(
    place       // MessagePlace-Instanz, auf der die Meldung exklusiv angezeigt werden soll
) {
    this.remove();
    if (place instanceof MessagePlace) {
        place.register(this);
        this._place = place;
    } else
        this._place = null;
}

/**
* Diese Meldung von der Anzeigefläche entfernen
* @return undefined
*/
Message.prototype.remove = function() {
    if (this._place)
        this._place.unregister(this);
    this._place = null;
}

/**
* Platz für Meldungen
*/
function MessagePlace(
    multiple,       // Optional: Sollen mehrere Meldungen angezeigt werden können?
    scrollIntoView  // Optional: Sollen neue Nachrichten automatisch in den sichtbaren Bereich gescrollt werden?
) {
    Object.defineProperty(this, 'node', {
        enumerable: true,
        value: createHtml('div', null, { 'class': 'messagePlace' })
    });
    this.messages = [];
    if (multiple)
        this.multiple = true;
    if (scrollIntoView)
        this.scrollIntoView = true;
}

// Angezeigte Meldungen
MessagePlace.prototype.messages = [];

// Sollen mehrere Meldungen angezeigt werden können?
MessagePlace.prototype.multiple = false;

// Sollen neue Nachrichten automatisch in den sichtbaren Bereich gescrollt werden?
MessagePlace.prototype.scrollIntoView = false;

// Sind Nachrichten sichtbar oder verborgen?
Object.defineProperty(MessagePlace.prototype, 'visible', { 'enumerable': true,
    get() { return !this.node.hasAttribute('style'); },
    set(v) {
        if (v != this.node.hasAttribute('style'))
            return;
        if (v)
            this.node.removeAttribute('style');
        else
            this.node.setAttribute('style', 'display:none');
    }
});

/**
* Platz für Meldungen im DOM-Baum vor nodeAfter einfügen
* @return undefined
*/
MessagePlace.prototype.insertThisBefore = function(
    nodeAfter       // Knoten, vor dem eingefügt werden soll
) {
    nodeAfter.parentNode.insertBefore(this.node, nodeAfter);
}

/**
* Platz für Meldungen im DOM-Baum nach nodeBefore einfügen
* @return undefined
*/
MessagePlace.prototype.insertThisAfter = function(
    nodeBefore      // Knoten, nach dem eingefügt werden soll
) {
    nodeBefore.parentNode.insertBefore(this.node, nodeBefore.nextSibling);
}

/**
* Platz für Meldungen im DOM-Baum als letztes Kind von parentNode einfügen
* @return undefined
*/
MessagePlace.prototype.appendAsChild = function(
    parentNode      // Künftiger Elternknoten
) {
    parentNode.appendChild(this.node);
}

/**
* Message-Instanz msg als eigene Nachricht registrieren und in DOM-Baum einfügen
* @return undefined
*/
MessagePlace.prototype.register = function(
    msg         // Zu registerende Message-Instanz
) {
    if (!(msg instanceof Message))
        return;
    this.node.appendChild(msg.node);
    this.messages.push(msg);
    if (this.scrollIntoView && (typeof msg.node.scrollIntoView == 'function'))
        msg.node.scrollIntoView({ behavior: 'smooth' });
}

/**
* Message-Instanz msg von ihrer MessagePlace-Instanz und aus dem DOM-Baum entfernen
* @return undefined
*/
MessagePlace.prototype.unregister = function(
    msg         // Zu entferndende Message-Instanz
) {
    if (!(msg instanceof Message))
        return;
    this.messages = this.messages.filter(function(o) { return (o != msg); });
    if (msg.node.parentNode)
        msg.node.parentNode.removeChild(msg.node);
}

/**
* Alle aktuell angezeigten Meldungen entfernen
* @return undefined
*/
MessagePlace.prototype.clear = function() {
    this.messages.map(function(msg) { msg.remove(); });
    this.messages = [];
}

/**
* Eine Message-Instanz msg anzeigen
* @return undefined
*/
MessagePlace.prototype.append = function(
    msg         // Anzuzeigende Message-Instanz
) {
    if (!(msg instanceof Message))
        return;
    if (!this.multiple)
        this.clear();
    msg.addTo(this);
}

/**
* Eine Meldung anlegen und anzeigen
* @return Message   Erstellte Meldung
*/
MessagePlace.prototype.add = function(
    text,       // Meldungstext
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    var msg = new Message(text, type);
    this.append(msg);
    return msg;
}

/** Anzeige für grundlegenden Lernpaketstatus (wie erfolgreich/erfolglos erledigt, Verbindungsprobleme etc.) */
function GeneralStateDisplay() {
    this.msgPlace = new MessagePlace();
    this.messages = Object.create(this.messages);
    this.messageTypes = Object.create(this.messageTypes);
}

/** Genutzte MessagePlace-Instanz */
GeneralStateDisplay.prototype.msgPlace = null;

/** ScoStateMgr-Instanz, die Lernpaketstatus speichert */
GeneralStateDisplay.prototype.ssm = null;

/** Anzuzeigende Meldungen (je String oder Funktion, die String liefert) */
GeneralStateDisplay.prototype.messages = {
    "NO CONNECTION": "Verbindungsaufnahme zur Lernplattform fehlgeschlagen - Fortschritte können nicht gespeichert werden",
    "OVERFLOW": "Eingabe ist zu lang und kann daher nicht gespeichert werden",
    "IDLE": "",
    "FAILED": "Aufgabe kann nicht mehr gelöst werden",
    "DONE": function() {
        const fine = this.ssm ? this.ssm.getValue('content.fine') : undefined;
        if (!Number.isFinite(fine))
            return "Aufgabe erfolgreich abgeschlossen";
        if (!fine)
            return 'Aufgabe fehlerfrei gelöst';
        if (fine == 1)
            return 'Aufgabe mit einem Fehler gelöst';
        return 'Aufgabe mit ' + fine + ' Fehlern gelöst';
    }
};

/** Zum jeweiligen Eintrag in this.messages gehörende Art der Meldung (vgl. Message.TYPE_*) */
GeneralStateDisplay.prototype.messageTypes = {
    "NO CONNECTION": Message.TYPE_WARN,
    "OVERFLOW": Message.TYPE_WARN,
    "IDLE": Message.TYPE_INFO,
    "FAILED": Message.TYPE_FAILED,
    "DONE": Message.TYPE_SUCCESS
};

/** Meldung mit internem Namen `key` anzeigen */
GeneralStateDisplay.prototype.showMessage = function(
    key     // Interner Name der Meldung (vgl. Attributnamen von `this.messages` und `this.messageTypes`)
) {
    const text = (typeof this.messages[key] == 'function') ? this.messages[key].call(this) : this.messages[key];
    if (((typeof text != 'string') && !(text instanceof String)) || !text.trim().length) {
        this.msgPlace.clear();
        return;
    }
    this.msgPlace.add(text, this.messageTypes[key]);
}

/**
* Platz für Meldungen im DOM-Baum vor nodeAfter einfügen
* @return undefined
*/
GeneralStateDisplay.prototype.insertThisBefore = function(
    nodeAfter       // Knoten, vor dem eingefügt werden soll
) {
    this.msgPlace.insertThisBefore(nodeAfter);
}

/**
* Platz für Meldungen im DOM-Baum nach nodeBefore einfügen
* @return undefined
*/
GeneralStateDisplay.prototype.insertThisAfter = function(
    nodeBefore      // Knoten, nach dem eingefügt werden soll
) {
    this.msgPlace.insertThisAfter(nodeBefore);
}

/**
* Platz für Meldungen im DOM-Baum als letztes Kind von parentNode einfügen
* @return undefined
*/
GeneralStateDisplay.prototype.appendAsChild = function(
    parentNode      // Künftiger Elternknoten
) {
    this.msgPlace.appendAsChild(parentNode);
}

/**
* Neu festlegen, welche Meldung aktuell angezeigt wird
* @return undefined
*/
GeneralStateDisplay.prototype.update = function() {
const connectedLms = this.ssm ? this.ssm.getValue('connection.lms') : '?';
const connectedBroadcast = this.ssm ? this.ssm.getValue('connection.broadcast') : '?';
    if (!this.ssm || !(this.ssm.getValue('connection.lms') || this.ssm.getValue('connection.broadcast')))
        return this.showMessage("NO CONNECTION");
    if (this.ssm.getValue('pool4lms.unsatisfied') >  ScoStateMgr.PRIORITY_HOPEDFOR)
        return this.showMessage("OVERFLOW");
    switch (this.ssm.getValue('lms.lessonStatus')) {
        case 'passed':
        case 'completed':
            return this.showMessage("DONE");
        case 'failed':
            return this.showMessage("FAILED");
    }
    return this.showMessage("IDLE");
}

/**
* ScoStateMgr-Instanz festlegen, die genutzt werden soll
* @return undefined
*/
GeneralStateDisplay.prototype.useScoStateMgr = function(
    ssm     // Zu nutzende ScoStateMgr-Instanz
) {
    this.ssm = ssm;
    ssm.addListener('connection.lms', () => this.update(), false);
    ssm.addListener('connection.broadcast', () => this.update(), false);
    ssm.addListener('pool4lms.unsatisfied', () => this.update(), false);
    ssm.addListener('lms.lessonStatus', () => this.update(), false);
    this.update();
}

/**
* Klasse, die einen zusätzlichen Abschnitt im Footer verwaltet
*/
function FooterSection() {
    FooterSection.object = this;
    var footer = document.getElementsByTagName('footer');
    if (footer.length)
        footer = footer[footer.length - 1];
    else
        footer = createHtml('footer', document.body);
    this.node = createHtml('span', footer, null, undefined, footer.firstChild);
}

/**
* Bisherige FooterSection-Instanz aufrufen oder eine FooterSection-Instanz erzeugen, wenn es noch keine gibt
* @return FooterSection     FooterSection-Instanz
*/
FooterSection.getObject = function() {
    if (!FooterSection.object)
        FooterSection.object = new FooterSection();
    return FooterSection.object;
}

// DOM-Repräsentation
FooterSection.prototype.node = null;

/**
* Die ImageLink-Instanz oder den DOM-Knoten obj aufnehmen
* @return undefined
*/
FooterSection.prototype.appendChild = function(
    obj         // Aufzunehmende ImageLink-Instanz oder DOM-Knoten
) {
    if (obj instanceof ImageLink)
        obj = obj.node;
    this.node.appendChild(obj);
    this.node.appendChild(document.createTextNode(' '));
}

/**
* Die ImageLink-Instanz oder den DOM-Knoten obj entfernen
* @return undefined
*/
FooterSection.prototype.removeChild = function(
    obj         // Aufzunehmende ImageLink-Instanz oder DOM-Knoten
) {
    if (obj instanceof ImageLink)
        obj = obj.node;
    if (obj.parentNode != this.node)
        return;
    var next = obj.nextSibling;
    this.node.removeChild(obj);
    if (next)
        this.node.removeChild(next);
}

/**
* Jede Instanz dieser Klasse ist ein Objekt, dass zu Beginn definierten Objekten mitteilt, wenn eins ihrer vorgegebenen Attribute sich ändert
*/
function AttributeObservableObject(
    attributes,     // Array mit Informationen über die Attribute in der Form { name: 'Attributname', default: 'Standardwert', … }
    observers       // Array mit Objekten, deren Methode registerDataChange(attributename, values) aufgerufen werden soll, wenn sich ein Wert ändert, sofern deren Methode isRelevantData(attributeInfo) den Wert true zurückgibt; registriert sich bei den Objekten durch Aufruf der Methode registerAttributeObservableObject(this)
) {
    const values = {};
    const self = this;
    for (let n = 0; n < attributes.length; ++n) {
        values[attributes[n].name] = attributes[n].default;
        const getter = (attr => (() => values[attr]))(attributes[n].name);
        const setter = ((attr, observers) => (v => {
            try {
                if (JSON.stringify(values[attr]) == JSON.stringify(v))
                    return;
            } catch(e) {
                return;
            }
            values[attr] = v;
            observers.forEach(obj => obj.registerDataChange(attr, self));
        }))(attributes[n].name, observers.filter(observer => observer.isRelevantData(attributes[n])));
        Object.defineProperty(this, attributes[n].name, { enumerable: true, set: setter, get: getter })
    }
    observers.forEach(observer => observer.registerAttributeObservableObject(this));
}


ScoTools.ScoStateMgr = ScoStateMgr;
ScoTools.createNS = createNS;
ScoTools.createHtml = createHtml;
ScoTools.createSvg = createSvg;
ScoTools.ModalPage = ModalPage;
ScoTools.InfoPageLink = InfoPageLink;
ScoTools.ImageLink = ImageLink;
ScoTools.Message = Message;
ScoTools.AttributeObservableObject = AttributeObservableObject;
Object.defineProperty(ScoTools, 'MSGTYPE_SUCCESS', { enumerable: true, value: Message.TYPE_SUCCESS });
Object.defineProperty(ScoTools, 'MSGTYPE_FAILED', { enumerable: true, value: Message.TYPE_FAILED });
Object.defineProperty(ScoTools, 'MSGTYPE_INFO', { enumerable: true, value: Message.TYPE_INFO });
Object.defineProperty(ScoTools, 'MSGTYPE_WARN', { enumerable: true, value: Message.TYPE_WARN });
Object.defineProperty(ScoTools, 'MSGTYPES', { enumerable: true, value: Message.TYPES });
ScoTools.MessagePlace = MessagePlace;
ScoTools.GeneralStateDisplay = GeneralStateDisplay;
ScoTools.FooterSection = FooterSection;
ScoTools.getFooterSection = FooterSection.getObject;
return ScoTools;
})();
window.createHtml = ScoTools.createHtml;

